# -*- coding: utf-8 -*-

"""External IBM Watson web application"""

__all__ = ['IBMWatsonApp']

import typing
from anaconda_navigator.static import images
from . import base

if typing.TYPE_CHECKING:
    from anaconda_navigator.config import user as user_config


class IBMWatsonApp(base.BaseWebApp):
    """Shortcut for IBM Watson web application."""

    def __init__(self, config: 'user_config.UserConfig') -> None:
        super().__init__(
            app_name='ibm_watson',
            display_name='IBM Watson Studio Cloud',
            description=(
                'IBM Watson Studio Cloud provides you the tools to analyze and visualize data, to cleanse and shape '
                'data, to create and train machine learning models. Prepare data and build models, using open source '
                'data science tools or visual modeling.'
            ),
            image_path=images.IBM_WATSON_ICON_PATH,
            url='http://www.anaconda.com/ibm_wsc_navigator',
            is_available=True,
            needs_license=False,
            config=config,
        )
